
/*
 * option.c
 */
#include "config.h"
#include "copyright.h"

#include <stdio.h>
#include <ctype.h>
#include "Wlib.h"
#include "defs.h"
#include "struct.h"
#include "data.h"
#include "packets.h"

/* forward decl */
int optionaction ();

int notdone;			/* not done flag */

#ifdef ROTATERACE
static int old_rotate, old_rotate_deg;
#endif

static int lastUpdateSpeed = 5;
static char newkeys[14];

#if (defined( DEBUG) || defined (BITMAP_DEBUG)) && defined(DYNAMIC_BITMAPS)
extern int OwnBitmapNum;
#endif

char *localmes[] =
{"Show owner on local planets",
 "Show resources on local planets",
 "Show nothing on local planets",
 ""};

char *galacticmes[] =
{"Show owner on galactic map",
 "Show resources on galactic map",
 "Show nothing on galactic map",
 ""};

#ifdef ROTATERACE
char *rotatemess[] =
{"Don't rotate galaxy",
 "Rotate galaxy 90 degrees",
 "Rotate galaxy 180 degrees",
 "Rotate galaxy 270 degrees",
 ""
};
#endif

char *mapupdates[] =
{"Don't update galactic map",
 "Update galactic map rarely",
 "Update galactic map frequently",
 ""};

static char *lockoptions[] =
{"Don't show lock icon",
 "Show lock icon on galactic map only",
 "Show lock icon on tactical map only",
 "Show lock icon on both map windows", ""};

/* useful for options that are an int with a range */
struct int_range
  {
    int min_value;		/* value is >= this */
    int max_value;		/* value is <= this */
    int increment;		/* a click raises/lowers this amount */
  };


/*
 * Only one of op_option, op_targetwin, and op_string should be
 * defined. If op_string is defined, op_size should be too and
 * op_text is used without a "Don't" prefix.
 * if op_range is defined, there should be a %d in op_text for it, op_size
 * will be non-useful, and the 'Don't ' prefix won't appear
 */
struct option
  {
    int op_num;
    char *op_text;		/* text to display when on */
    int *op_option;		/* variable to test/modify (optional) */
    W_Window *op_targetwin;	/* target window to map/unmap (optional) */
    char *op_string;		/* string to modify (optional) */
    int op_size;		/* size of *op_string (optional) */
    char **op_array;		/* array of strings to switch between */
    struct int_range *op_range;	/* struct definint an integer range option */
  };
static char *phaseroptions[] =
{"Don't log phaser hits",
 "log phasers on kill window",
 "log phasers on phaser window",
 "log phasers on review window only", ""};

/* for the paged options menus */
struct option_menu
  {
    int page_num;		/* page number of this menu */
    struct option_menu *Next;
    struct option *menu;	/* pointers to arrary of options */
    int numopt;			/* number of options in this menu page */
    int updated;		/* 1 if options can be changed externally */
  };

/* pointer to first entry in the options menu list */
struct option_menu *FirstMenu = NULL;
struct option_menu *CurrentMenu = NULL;	/* menu currently looked at */
int MenuPage = 0;		/* current menu page */
int MaxOptions = 0;		/* maximum number of options in all menu pages */
struct int_range MenuPages =
{0, 1, 1};


/* updates: use of the int range thing... */
struct int_range updates_range =
{1, 10, 1};
#ifdef RABBIT_EARS
struct int_range rabbit_range =
{0, 2, 1};
#endif
/* range of menus. Will be updated when menu list is assembled */
struct int_range Menus_Range =
{0, 1, 1};

#if (defined( DEBUG) || defined (BITMAP_DEBUG)) && defined(DYNAMIC_BITMAPS)
struct int_range bitmap_range =
{0, 50, 1};
#endif

struct int_range keepInfo_range =
{0, 100, 1};

/* menus */
struct option Window_Menu[] =
{
  {0, "Window Menu", &MenuPage, 0, 0, 0, NULL, &Menus_Range},
  {1, "Page %d (click to change)", &MenuPage, 0, 0, 0, NULL, &Menus_Range},
  {2, "show \"all\" message window", 0, &messwa, 0, 0, NULL, NULL},
  {3, "show \"team\" message window", 0, &messwt, 0, 0, NULL, NULL},
  {4, "show \"your\" message window", 0, &messwi, 0, 0, NULL, NULL},
  {5, "show \"kill\" message window", 0, &messwk, 0, 0, NULL, NULL},
  {6, "show \"total\" message window", 0, &reviewWin, 0, 0, NULL, NULL},
  {7, "show phaser log window", &phaserWindow, &phaserwin, 0, 0, NULL},
  {8, "show statistic window", 0, &statwin, 0, 0, NULL, NULL},
  {9, "show ping stats window", 0, &pStats, 0, 0, NULL},
  {10, "show network statistic window", 0, &netstatWin, 0, 0, NULL, NULL},
  {11, "show LagMeter ", 0, &lMeter, 0, 0, NULL, NULL},
  {12, "show UDP control window", 0, &udpWin, 0, 0, NULL, NULL},
#ifdef SHORT_PACKETS
  {13, "show Short Packages window", 0, &spWin, 0, 0, NULL, NULL},
#endif
  {14, "show help window", 0, &helpWin, 0, 0, NULL, NULL},
#ifdef XTREKRC_HELP
  {15, "show xtrekrc defaults window", 0, &defWin, 0, 0, NULL, NULL},
#endif
#ifdef DOC_WIN
  {16, "show xtrekrc", 0, &xtrekrcwin, 0, 0, NULL, NULL},
  {17, "show documentation window", 0, &docwin, 0, 0, NULL, NULL},
#endif
#ifdef TOOLS
  {18, "show shell tools window", 0, &toolsWin, 0, 0, NULL, NULL},
#endif
#ifdef SOUND
  {19, "show sound control window", 0, &soundWin, 0, 0, NULL, NULL},
#endif
  {20, "done", &notdone, 0, 0, 0, NULL, NULL},
  {-1, NULL, 0, 0, 0, 0, NULL, NULL}
};

struct option Features_Menu[] =
{
  {0, "Defaults Menu", &MenuPage, 0, 0, 0, NULL, &Menus_Range},
  {1, "Page %d (click to change)", &MenuPage, 0, 0, 0, NULL, &Menus_Range},
  {2, "show tactical planet names", &namemode, 0, 0, 0, NULL, NULL},
  {3, "show shields", &showShields, 0, 0, 0, NULL, NULL},
  {4, "stay peaceful when reborn", &keeppeace, 0, 0, 0, NULL, NULL},
  {5, "new keymap entries: %s", 0, 0, newkeys, 13, NULL, NULL},
  {6, "", &showlocal, 0, 0, 0, localmes, NULL},
  {7, "", &showgalactic, 0, 0, 0, galacticmes, NULL},
  {8, "%d updates per second", &updatespeed, 0, 0, 0, 0, &updates_range},
  {9, "avoid message kludge", &niftyNewMessages, 0, 0, 0, NULL, NULL},
  {10, "show tractor/pressor", &showTractorPressor, 0, 0, 0, NULL, NULL},
#ifdef VARY_HULL
  {11, "warn hull", &vary_hull, 0, 0, 0, NULL, NULL},
#endif
#ifdef HOCKEY_LINES
  {12, "show hockey lines", &hockey_s_lines, 0, 0, 0, NULL, NULL},
#endif
  {13, "done", &notdone, 0, 0, 0, NULL, NULL},
  {-1, NULL, 0, 0, 0, 0, NULL, NULL}
};

struct int_range enPhasRng =
{0, 10, 1};

struct option SillyFeatures_Menu[] =
{
  {0, "Extra Features Menu", &MenuPage, 0, 0, 0, NULL, &Menus_Range},
  {1, "Page %d (click to change)", &MenuPage, 0, 0, 0, NULL, &Menus_Range},
  {2, "", &showLock, 0, 0, 0, lockoptions, NULL},
  {3, "sort players in player window", &sortPlayers, 0, 0, 0, NULL, NULL},
{4, "show tractor/pressor beams", &showTractorPressor, 0, 0, 0, NULL, NULL},
  {5, "show tractors after lock ", &continuetractor, 0, 0, 0, NULL, NULL},
#ifdef RABBIT_EARS
  {6, "Use silly planet bitmaps %d", &myPlanetBitmap, 0, 0, 0, NULL, &rabbit_range},
#else
  {6, "Use silly planet bitmaps", &myPlanetBitmap, 0, 0, 0, NULL, NULL},
#endif
  {7, "Show planet owner", &showPlanetOwner, 0, 0, 0, NULL, NULL},
  {7, "alert on extra border(s)", &extraBorder, 0, 0, 0, NULL, NULL},
#ifdef ROTATERACE
  {8, "", &rotate, 0, 0, 0, rotatemess, NULL},
#endif

#if (defined( DEBUG) || defined (BITMAP_DEBUG)) && defined(DYNAMIC_BITMAPS)
  {9, "Own bitmap number: %d", &OwnBitmapNum, 0, 0, 0, NULL, &bitmap_range},
#endif

#ifdef VSHIELD_BITMAPS
  {9, "vary shields bitmap", &VShieldBitmaps, 0, 0, 0, NULL, NULL},
#else
  {9, "use warning Shields", &warnShields, 0, 0, 0, NULL, NULL},
#endif

  {10, "use ROMVLVS bitmap", &ROMVLVS, 0, 0, 0, NULL, NULL},
#ifdef TNG_FED_BITMAPS
  {10, "use fed TNG bitmaps", &use_tng_fed_bitmaps, 0, 0, 0, NULL, NULL},
#endif
  {8, "keep info %d upds (0=don't remove)", &keepInfo, 0, 0, 0, 0, &keepInfo_range},
  {10, "use new dashboard", &newDashboard, 0, 0, 0, NULL, NULL},
  {10, "use new distress", &UseNewDistress, 0, 0, 0, NULL, NULL},
  {10, "use new player list", &newPlist, 0, 0, 0, NULL, NULL},
  {10, "show IND planets", &showIND, 0, 0, 0, NULL, NULL},
  {10, "enemy phaser width: %d", &enemyPhasers, 0, 0, 0, 0, &enPhasRng},
  {14, "done", &notdone, 0, 0, 0, NULL, NULL},
  {-1, NULL, 0, 0, 0, 0, NULL, NULL}
};
#define NUMOPTIONS(menu) ((sizeof((menu))/sizeof((menu)[0]))-1)

/* option menu sizes and such */
#define OPTIONBORDER	2
#define OPTIONLEN	35

/* Set up the option menus and window. */
optionwindow ()
{
  register int i;

  /* Init not done flag */
  notdone = 1;

  *newkeys = '\0';
  if (FirstMenu == NULL)
    {
      MaxOptions = InitOptionMenus ();
      if (MaxOptions < 0)
	{
	  fprintf (stderr, "InitOptionMenus() error %s!\n", MaxOptions);
	  notdone = 0;
	  return;
	}
    }

  /* Create window big enough to hold option windows */
  if (optionWin == NULL)
    {

      optionWin = W_MakeMenu ("option", WINSIDE + 10, -BORDER + 10, OPTIONLEN,
			      MaxOptions, baseWin, OPTIONBORDER);
      W_SetWindowKeyDownHandler (optionWin, optionaction);
      W_SetWindowButtonHandler (optionWin, optionaction);
      W_DefineArrowCursor (optionWin);
      CurrentMenu = FirstMenu;

      RefreshOptions ();
    }

  /* Map window */
  W_MapWindow (optionWin);
}

/* refresh all current options */
RefreshOptions ()
{
  int i;
  struct option_menu *option;

  if (notdone == 0 || (option = CurrentMenu) == NULL)
    return;

  for (i = 0; i < option->numopt; i++)
    {
      optionrefresh (&(option->menu[i]));
    }
  if (option->numopt < MaxOptions)
    for (i = option->numopt; i < MaxOptions; i++)
      {
	OptionClear (i);
      }
}

/* blank out option line 'i' */
OptionClear (i)
{
  char *blanktext = "                                               ";
  if (optionWin && notdone)
    W_WriteText (optionWin, 0, i, textColor, blanktext, OPTIONLEN, 0);
}

/* Redraw the specified option entry */
optionredrawtarget (win)
     W_Window win;
{
  register struct option *op;

  if (notdone == 0)
    return;

  for (op = CurrentMenu->menu; op->op_text; op++)
    {
      if (op->op_targetwin && win == *op->op_targetwin)
	{
	  optionrefresh (op);
	  break;
	}
    }
}

/* Redraw the specified option option */
optionredrawoption (ip)
     int *ip;
{
  register struct option *op;

  if (notdone == 0)
    return;

  for (op = CurrentMenu->menu; op->op_num >= 0; op++)
    {
      if (ip == op->op_option)
	{
	  optionrefresh (op);
	  break;
	}
    }
}

/* Refresh the option window given by the option struct */
optionrefresh (op)
     register struct option *op;
{
  register int on;
  char buf[BUFSIZ];

  if (op == NULL || notdone == 0)
    return;

  if (op->op_string)
    {
      (void) sprintf (buf, op->op_text, op->op_string);
    }
  else if (op->op_array)
    {				/* Array of strings */
      strcpy (buf, op->op_array[*op->op_option]);
    }
  else if (op->op_range)
    {
      (void) sprintf (buf, op->op_text, *(op->op_option));
    }
  else
    {
      /* Either a boolean or a window */
      if (op->op_option)
	on = *op->op_option;	/* use int for status */
      else if (op->op_targetwin)
	on = W_IsMapped (*op->op_targetwin);	/* use window for status */
      else
	on = 1;			/* shouldn't happen */

      if (!on)
	strcpy (buf, "Don't ");
      else
	buf[0] = '\0';
      strcat (buf, op->op_text);
    }

  if (islower (buf[0]))
    buf[0] = toupper (buf[0]);

  W_WriteText (optionWin, 0, op->op_num, textColor, buf, strlen (buf), 0);
}

/* deal with events sent to the option window */
optionaction (data)
     W_Event *data;
{
  register struct option *op;
  int i;
  register char *cp;

  if (data->y >= CurrentMenu->numopt)
    {
      W_Beep ();
      return (0);
    }
  if (notdone == 0)
    return (0);

  op = &(CurrentMenu->menu[data->y]);

  /* Update string; don't claim keystrokes for non-string options */
  /* deal with options with string input first */
  if (op->op_string == 0)
    {
      if (data->type == W_EV_KEY)
	return (0);
    }
  else
    {
      if (data->type == W_EV_BUTTON)
	return (0);
      switch (data->key)
	{

	case '\b':		/* delete character */
	case '\177':
	  cp = op->op_string;
	  i = strlen (cp);
	  if (i > 0)
	    {
	      cp += i - 1;
	      *cp = '\0';
	    }
	  break;

	case '\027':		/* word erase */
	  cp = op->op_string;
	  i = strlen (cp);
	  /* back up over blanks */
	  while (--i >= 0 && isspace (cp[i]))
	    ;
	  i++;
	  /* back up over non-blanks */
	  while (--i >= 0 && !isspace (cp[i]))
	    ;
	  i++;
	  cp[i] = '\0';
	  break;

	case '\025':		/* kill line */
	case '\030':
	  op->op_string[0] = '\0';
	  break;

	default:		/* add character to the list */
	  if (data->key < 32 || data->key > 127)
	    break;
	  cp = op->op_string;
	  i = strlen (cp);
	  if (i < (op->op_size - 1) && !iscntrl (data->key))
	    {
	      cp += i;
	      cp[1] = '\0';
	      cp[0] = data->key;
	    }
	  else
	    W_Beep ();
	  break;
	}
    }

  /* Toggle int, if it exists */
  if (op->op_array)
    {
      if (data->key == W_RBUTTON)
	{
	  (*op->op_option)++;
	  if (*(op->op_array)[*op->op_option] == '\0')
	    {
	      *op->op_option = 0;
	    }
	}
      else if (data->key == W_MBUTTON)
	{
	  /* set option number to zero on the middle key to ease shutoff */
	  *op->op_option = 0;
	}
      else if (data->key == W_LBUTTON)
	{
	  /* if left button, decrease option  */
	  (*op->op_option)--;
	  /* if decreased too far, set to top option */
	  if (*(op->op_option) < 0)
	    {
	      *op->op_option = 0;
	      while (*(op->op_array)[*op->op_option] != '\0')
		{
		  (*op->op_option)++;
		}
	      (*op->op_option)--;
	    }
	}

#ifdef ROTATERACE
      if (rotate != old_rotate)
	{
	  register i;
	  register struct planet *l;
	  register struct player *j;

	  redrawall = 1;
	  reinitPlanets = 1;

	  for (i = 0, l = planets; i < MAXPLANETS; i++, l++)
	    {
	      if (rotate)
		{
		  rotate_deg = -old_rotate_deg + rotate * 64;
		  rotate_coord (&l->pl_x, &l->pl_y, rotate_deg, GWIDTH / 2, GWIDTH / 2);
		  rotate_deg = rotate * 64;
		}
	      else
		{
		  rotate_deg = -old_rotate_deg;
		  rotate_coord (&l->pl_x, &l->pl_y, rotate_deg, GWIDTH / 2, GWIDTH / 2);
		  rotate_deg = 0;
		}
	    }

	  /* we could wait for the server to do this but looks better if we
	    do it now. */
	  for (i = 0, j = players; i < MAXPLAYER; i++, j++)
	    {
	      if (j->p_status != PALIVE)
		continue;
	      if (rotate)
		{
		  rotate_deg = -old_rotate_deg + rotate * 64;
		  rotate_coord (&j->p_x, &j->p_y, rotate_deg,
				GWIDTH / 2, GWIDTH / 2);
		  rotate_dir (&j->p_dir, rotate_deg);

		  rotate_deg = rotate * 64;
		}
	      else
		{
		  rotate_deg = -old_rotate_deg;
		  rotate_coord (&j->p_x, &j->p_y, rotate_deg,
				GWIDTH / 2, GWIDTH / 2);
		  rotate_dir (&j->p_dir, rotate_deg);
		  rotate_deg = 0;
		}
	    }
	  /* phasers/torps/etc .. wait for server */

	  old_rotate = rotate;
	  old_rotate_deg = rotate_deg;
	}
#endif
    }
  else if (op->op_range)
    {
      if (data->key == W_RBUTTON)
	{
	  (*op->op_option) += op->op_range->increment;
	}
      else if (data->key == W_MBUTTON)
	{
	  (*op->op_option) = op->op_range->min_value;
	}
      else if (data->key == W_LBUTTON)
	{
	  (*op->op_option) -= op->op_range->increment;
	}
      /* wrap value around within option range */
      if (*(op->op_option) > op->op_range->max_value)
	*(op->op_option) = op->op_range->min_value;
      if (*(op->op_option) < op->op_range->min_value)
	*(op->op_option) = op->op_range->max_value;
    }
  else if (op->op_option)
    {
      *op->op_option = !*op->op_option;
    }

  /* Map/unmap window, if it exists */
  if (op->op_targetwin)
    {
      if (W_IsMapped (*op->op_targetwin))
	W_UnmapWindow (*op->op_targetwin);
      else
	{
	  W_MapWindow (*op->op_targetwin);

	  if (*op->op_targetwin == udpWin)
	    udpwindow ();
	  if (*op->op_targetwin == pStats)
	    redrawPStats ();
	  if (*op->op_targetwin == netstatWin)
	    nswindow ();
#ifdef XTREKRC_HELP
	  if (defWin && *op->op_targetwin == defWin)
	    showdef ();
#endif
#ifdef SOUND
	  if (*op->op_targetwin == soundWin)
	    soundwindow ();
#endif
#ifdef DOC_WIN
	  if (docwin && *op->op_targetwin == docwin)
	    showdocs (0);

	  if (xtrekrcwin && *op->op_targetwin == xtrekrcwin)
	    showxtrekrc (0);
#endif
	}
    }

  /* deal with possible menu change */
  if (MenuPage != CurrentMenu->page_num)
    {
      SetMenuPage (MenuPage);
      RefreshOptions ();
    }

  if (notdone)
    optionrefresh (op);
  else
    {
      optionrefresh (op);
      optiondone ();
    }

  return (1);
}

/* find the menu in the menus linked list that matches the one in the
   ** argument */
SetMenuPage (pagenum)
     int pagenum;
{
  int i = 1;
  if (FirstMenu != NULL)
    for (CurrentMenu = FirstMenu; CurrentMenu->Next != NULL &&
    CurrentMenu->page_num != pagenum; i++, CurrentMenu = CurrentMenu->Next);
}

optiondone ()
{
  char *str;

  /* Unmap window */
  W_UnmapWindow (optionWin);

  /* update keymap */
  for (str = newkeys; *str != '\0'; str += 2)
    {
      if ((*str >= 32 && *str < 127) || *str == 'O')
	{
	  if (*(str + 1) == '\0')
	    break;
	  mystats->st_keymap[*str - 32] = *(str + 1);
	}
      if (*(str + 1) == '\0')
	break;
    }
  *newkeys = '\0';

  if (updatespeed != lastUpdateSpeed)
    {
      sendUpdatePacket (1000000 / updatespeed);
      lastUpdateSpeed = updatespeed;
    }

  sendOptionsPacket ();		/* update server as to the client's options */
}

/* set up menus linked list */
int
InitOptionMenus ()
{
  int i = 1;
  int maxopts = 0;


  /* AddOptMenu( &OptionsMenu, 0); */
  AddOptMenu (Features_Menu, 0);
  AddOptMenu (Window_Menu, 0);
  AddOptMenu (SillyFeatures_Menu, 0);

  for (i = 0, CurrentMenu = FirstMenu; CurrentMenu != NULL;
       i++, CurrentMenu = CurrentMenu->Next)
    {
      CurrentMenu->page_num = i;/* repage the menus.. */
      if (CurrentMenu->numopt > maxopts)
	maxopts = CurrentMenu->numopt;
    }
  CurrentMenu = FirstMenu;
  Menus_Range.max_value = i - 1;
  return maxopts;
}

AddOptMenu (NewMenu, updated)
     struct option NewMenu[];
     int updated;
{
  struct option_menu *menuptr;
  struct option_menu *newmenu;
  int i = 0;


  menuptr = FirstMenu;

  newmenu = (struct option_menu *) malloc (sizeof (struct option_menu));
  if (newmenu == NULL)
    {
      perror ("Malloc Error adding a menu");
      return;
    }

  /* add to list */
  if (FirstMenu == NULL)
    {
      FirstMenu = newmenu;
    }
  else
    {
      for (i = 0, menuptr = FirstMenu; menuptr->Next != NULL; menuptr = menuptr->Next)
	i++;
      menuptr->Next = newmenu;
    }
  newmenu->page_num = i;
  newmenu->Next = NULL;
  newmenu->numopt = NumOptions (NewMenu);
  newmenu->menu = NewMenu;
  newmenu->updated = updated;
}

int
NumOptions (OpMenu)
     struct option OpMenu[];
{
  int i = 0;
  struct option *ptr;

  ptr = &OpMenu[0];
  for (i = 0; ptr->op_num != -1 && ptr->op_option != &notdone; i++)
    {
      ptr = &OpMenu[i];
      ptr->op_num = i;
    }


  return i;
}

/* a function that could be called regularly, to deal with menus that
   ** might be updated by external events. I.e. the udp menu!
 */
UpdateOptions ()
{
  if (notdone == 0)
    return;			/* don't update if menu isn't in use */
  if (CurrentMenu->updated)
    RefreshOptions ();
}
